from enum import Enum, StrEnum


class DiagramLayout(StrEnum):
    """Layout of a Mermaid diagram"""
    DAGRE = 'dagre'
    ELK = 'elk'


class DiagramLook(StrEnum):
    """Look of a Mermaid diagram"""
    NEO = 'neo'


class DiagramTheme(StrEnum):
    """Theme of a Mermaid diagram"""
    NEUTRAL = 'neutral'


class DiagramOrientation(StrEnum):
    """Orientation of a Mermaid graph"""
    TOP_DOWN = 'TD'
    BOTTOM_UP = 'BT'
    LEFT_RIGHT = 'LR'
    RIGHT_LEFT = 'RL'


class DiagramNodeShape(Enum):
    """Shapes of a Mermaid graph node"""
    DEFAULT = ("(", ")")
    RECTANGLE = ("[", "]")
    ROUND_RECTANGLE = ("([", "])")
    SUBROUTINE = ("[[", "]]")
    DATABASE = ("[(", ")]")
    CIRCLE = ("((", "))")
    DOUBLE_CIRCLE = ("(((", ")))")
    FLAG = (">", "]")
    DIAMOND = ("{", "}")
    HEXAGON = ("{{", "}}")
    PARALLELOGRAM = ("[/", "/]")
    PARALLELOGRAM_ALT = ("[\\", "\\]")
    TRAPEZOID = ("[/", "\\]")
    TRAPEZOID_ALT = ("[\\", "/]")



import threading
import webbrowser
from tempfile import TemporaryDirectory

import networkx as nx

from networkx_mermaid import DiagramNodeShape, DiagramOrientation
from networkx_mermaid.builders import DiagramBuilder
from networkx_mermaid.formatters import html, markdown
from networkx_mermaid.typing import MermaidDiagram


# An example of a graph with multiple components
def create_graph():
    pastel_colors = ["#FFCCCC", "#CCFFCC", "#CCCCFF", "#FFFFCC", "#CCFFFF", "#FFCCFF"]
    graphs: list[nx.Graph] = [nx.tetrahedral_graph(), nx.dodecahedral_graph()]

    for i, g in enumerate(graphs):
        nx.set_node_attributes(g, {n: {"color": pastel_colors[i]} for n in g.nodes})

    graph: nx.Graph = nx.disjoint_union_all(graphs)

    graph.name = " + ".join(g.name for g in graphs)

    return graph


def create_builder():
    # Create a Mermaid Diagram Builder with custom settings

    builder = DiagramBuilder(
        orientation=DiagramOrientation.LEFT_RIGHT,
        node_shape=DiagramNodeShape.ROUND_RECTANGLE,
    )
    return builder


def create_server(port: int, root_directory: str, open_browser: bool = True) -> threading.Thread:
    import http.server
    import socketserver

    url = f"http://localhost:{port}"

    class Handler(http.server.SimpleHTTPRequestHandler):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, directory=root_directory, **kwargs)

    def serve():
        with socketserver.TCPServer(('', port), Handler) as httpd:
            print("Serving at:", url)
            httpd.serve_forever()

    server_thread = threading.Thread(target=serve)
    server_thread.daemon = True
    server_thread.start()

    if open_browser:
        webbrowser.open(url)


def main():
    graph = create_graph()
    builder = create_builder()

    # Build the Mermaid Diagram
    mermaid_diagram: MermaidDiagram = builder.build(graph)

    # Format the Mermaid Diagram for Markdown embedding
    markdown_diagram: str = markdown(mermaid_diagram)

    # or as single page HTML
    html_diagram: str = html(mermaid_diagram, title=graph.name)

    print('Mermaid Diagram:')
    print(mermaid_diagram)
    print(markdown_diagram)
    print(html_diagram)

    ## Save the HTML diagram to a file and serve it
    with TemporaryDirectory() as temp_dir:
        with open(f"{temp_dir}/index.html", 'w') as f:
            f.write(html_diagram)

        # Serve the HTML diagram
        create_server(port=8073, root_directory=temp_dir, open_browser=True)

        # Keep the main thread alive to allow the server to run
        try:
            while True:
                pass
        except KeyboardInterrupt:
            print("Server stopped")


if __name__ == "__main__":
    main()



from networkx_mermaid.formatters import html, markdown


def test_markdown():
    # Arrange
    diagram = "graph LR\nA-->B"

    # Act
    result = markdown(diagram)

    # Assert
    assert result == "```mermaid\ngraph LR\nA-->B\n```"


def test_html_without_title():
    # Arrange
    diagram = "graph LR\nA-->B"
    expected_html_template = """<!doctype html>
<html lang="en">
  <head>
    <link rel="icon" type="image/x-icon" href="https://mermaid.js.org/favicon.ico">
    <meta charset="utf-8">
    <title>Mermaid Diagram</title>
    <style>
    pre.mermaid {
      font-family: "Fira Mono", "Roboto Mono", "Source Code Pro", monospace;
    }
    </style>
  </head>
  <body>
    <pre class="mermaid">
graph LR
A-->B
    </pre>
    <script type="module">
      import mermaid from 'https://cdn.jsdelivr.net/npm/mermaid@11/dist/mermaid.esm.min.mjs';
      let config = { startOnLoad: true, flowchart: { useMaxWidth: false, htmlLabels: true } };
      mermaid.initialize(config);
    </script>
  </body>
</html>
"""

    # Act
    result = html(diagram)

    # Assert
    assert result == expected_html_template


def test_html_with_title():
    # Arrange
    diagram = "graph LR\nA-->B"
    title = "My Diagram"
    expected_html_template = """<!doctype html>
<html lang="en">
  <head>
    <link rel="icon" type="image/x-icon" href="https://mermaid.js.org/favicon.ico">
    <meta charset="utf-8">
    <title>My Diagram</title>
    <style>
    pre.mermaid {
      font-family: "Fira Mono", "Roboto Mono", "Source Code Pro", monospace;
    }
    </style>
  </head>
  <body>
    <pre class="mermaid">
graph LR
A-->B
    </pre>
    <script type="module">
      import mermaid from 'https://cdn.jsdelivr.net/npm/mermaid@11/dist/mermaid.esm.min.mjs';
      let config = { startOnLoad: true, flowchart: { useMaxWidth: false, htmlLabels: true } };
      mermaid.initialize(config);
    </script>
  </body>
</html>
"""

    # Act
    result = html(diagram, title=title)

    # Assert
    assert result == expected_html_template



from . import builders, formatters, typing
from .builders import DiagramBuilder
from .models import DiagramLayout, DiagramLook, DiagramNodeShape, DiagramOrientation, DiagramTheme

__all__ = (
    'DiagramBuilder',
    'DiagramLayout',
    'DiagramLook',
    'DiagramNodeShape',
    'DiagramOrientation',
    'DiagramTheme',
    'builders',
    'formatters',
    'typing'
)



from .typing import MermaidDiagram

HTML_TEMPLATE = """<!doctype html>
<html lang="en">
  <head>
    <link rel="icon" type="image/x-icon" href="https://mermaid.js.org/favicon.ico">
    <meta charset="utf-8">
    <title>Mermaid Diagram</title>
    <style>
    pre.mermaid {
      font-family: "Fira Mono", "Roboto Mono", "Source Code Pro", monospace;
    }
    </style>
  </head>
  <body>
    <pre class="mermaid">
    </pre>
    <script type="module">
      import mermaid from 'https://cdn.jsdelivr.net/npm/mermaid@11/dist/mermaid.esm.min.mjs';
      let config = { startOnLoad: true, flowchart: { useMaxWidth: false, htmlLabels: true } };
      mermaid.initialize(config);
    </script>
  </body>
</html>
"""


def markdown(diagram: MermaidDiagram) -> str:
    """
    Generate Markdown code for a Mermaid diagram.
    """
    return f"```mermaid\n{diagram}\n```"


def html(diagram: MermaidDiagram, title: str | None = None) -> str:
    """
    Generate HTML code for a Mermaid diagram.
    """
    output = HTML_TEMPLATE.replace(
        '<pre class="mermaid">', f'<pre class="mermaid">\n{diagram}'
    )
    if title:
        output = output.replace(
            "<title>Mermaid Diagram</title>", f"<title>{title}</title>"
        )
    return output



from typing import Any

import networkx as nx
from mappingtools.collectors import MinifyingMapper

from .models import DiagramNodeShape, DiagramOrientation
from .typing import MermaidDiagram

DEFAULT_LAYOUT = "dagre"
DEFAULT_LOOK = "neo"
DEFAULT_THEME = "neutral"


def _edge_label(data: dict[str, Any]) -> str:
    """Generate an edge label string."""
    label = data.get("label")
    return f"|{label}|" if label else ""


def _contrast_color(color: str) -> str:
    """
    Return black or white by choosing the best contrast to input color.

    Args:
        color: str - hex color code

    Returns:
        color: str - hex color code
    """
    if not (isinstance(color, str) and color.startswith("#") and len(color) == 7):
        raise ValueError(f"Invalid color format: {color}. Expected a 6-digit hex code.")

    r, g, b = int(color[1:3], 16), int(color[3:5], 16), int(color[5:7], 16)
    return "#000000" if (r * 0.299 + g * 0.587 + b * 0.114) > 186 else "#ffffff"


def _node_style(node_id: str, data: dict[str, Any]) -> str:
    """Generate a node style string."""
    color = data.get("color")
    if color:
        return f"\nstyle {node_id} fill:{color}, color:{_contrast_color(color)}"
    return ""


def _graph_title(graph: nx.Graph, title: str | None = None) -> str:
    """Generate a graph title string."""
    title = title if title is not None else graph.name
    return f"title: {title}\n" if title else ""


class DiagramBuilder:
    """
    A class to generate Mermaid diagrams from NetworkX graphs.
    """

    def __init__(
            self,
            orientation: DiagramOrientation = DiagramOrientation.LEFT_RIGHT,
            node_shape: DiagramNodeShape = DiagramNodeShape.DEFAULT,
            layout: str = DEFAULT_LAYOUT,
            look: str = DEFAULT_LOOK,
            theme: str = DEFAULT_THEME,
    ):
        """
        Initialize the DiagramBuilder.

        Args:
            orientation: DiagramOrientation - The orientation of the graph (default: LEFT_RIGHT).
            node_shape: DiagramNodeShape - The shape of the nodes (default: DiagramNodeShape.DEFAULT).
            layout: str - the layout to use (default: 'dagre')
            look: str - the look to use (default: 'neo')
            theme: str - the theme to use (default: 'neutral')
        """
        self.orientation = orientation
        self.node_shape = node_shape
        self.layout = layout
        self.look = look
        self.theme = theme

        if not isinstance(orientation, DiagramOrientation):
            raise TypeError("orientation must be a valid Orientation enum")
        if not isinstance(node_shape, DiagramNodeShape):
            raise TypeError("node_shape must be a valid NodeShape enum")

    def _diagram_config(self, graph, title: str | None = None) -> str:
        return (
            f"---\n"
            f"{_graph_title(graph, title)}"
            f"config:\n"
            f"  layout: {self.layout}\n"
            f"  look: {self.look}\n"
            f"  theme: {self.theme}\n"
            f"---\n"
        )

    def build(self, graph: nx.Graph, title: str | None = None, with_edge_labels: bool = True) -> MermaidDiagram:
        """
        Materialize a graph as a Mermaid flowchart.

        Args:
            graph: nx.Graph - The NetworkX graph to convert.
            title: str - The title of the graph (default: None).
                   If None, the graph name will be used if available.
                   Supplying and empty string will remove the title.
            with_edge_labels: bool - Whether to include edge labels (default: True).

        Returns:
            A string representation of the graph as a Mermaid graph.
        """
        config = self._diagram_config(graph, title)

        bra, ket = self.node_shape.value

        minifier = MinifyingMapper()

        nodes = "\n".join(
            f"{minifier.get(u)}{bra}{d.get('label', u)}{ket}{_node_style(minifier.get(u), d)}" for u, d in
            graph.nodes.data())

        _edges = ((minifier.get(u), minifier.get(v), d) for u, v, d in graph.edges.data())
        edges = "\n".join(f"{u} -->{_edge_label(d) if with_edge_labels else ''} {v}" for u, v, d in _edges)

        return (
            f"{config}"
            f"graph {self.orientation.value}\n"
            f"{nodes}\n"
            f"{edges}"
        )



import networkx as nx
import pytest

from networkx_mermaid import DiagramNodeShape, DiagramOrientation
from networkx_mermaid.builders import (
    DEFAULT_LAYOUT,
    DEFAULT_LOOK,
    DEFAULT_THEME,
    DiagramBuilder,
    _contrast_color,
    _edge_label,
    _graph_title,
    _node_style,
)


def test_default_initialization():
    # Arrange
    builder = DiagramBuilder()

    # Act & Assert
    assert builder.orientation == DiagramOrientation.LEFT_RIGHT
    assert builder.node_shape == DiagramNodeShape.DEFAULT
    assert builder.layout == DEFAULT_LAYOUT
    assert builder.look == DEFAULT_LOOK
    assert builder.theme == DEFAULT_THEME


def test_custom_initialization():
    # Arrange
    builder = DiagramBuilder(
        orientation=DiagramOrientation.TOP_DOWN,
        node_shape=DiagramNodeShape.RECTANGLE,
        layout="elk",
        look="neo",
        theme="neutral"
    )

    # Act & Assert
    assert builder.orientation == DiagramOrientation.TOP_DOWN
    assert builder.node_shape == DiagramNodeShape.RECTANGLE
    assert builder.layout == "elk"
    assert builder.look == "neo"
    assert builder.theme == "neutral"


def test_invalid_orientation_type():
    # Arrange & Act & Assert
    with pytest.raises(TypeError):
        DiagramBuilder(orientation="invalid")


def test_invalid_node_shape_type():
    # Arrange & Act & Assert
    with pytest.raises(TypeError):
        DiagramBuilder(node_shape="invalid")


def test_build_empty_graph():
    # Arrange
    graph = nx.Graph()
    builder = DiagramBuilder()

    # Act
    diagram = builder.build(graph)

    # Assert
    assert "graph LR" in diagram


def test_build_simple_graph():
    # Arrange
    graph = nx.Graph()
    graph.add_edge(1, 2)
    builder = DiagramBuilder()

    # Act
    diagram = builder.build(graph)

    # Assert
    assert "graph LR" in diagram
    assert "A(1)" in diagram
    assert "B(2)" in diagram
    assert "A --> B" in diagram


def test_build_graph_with_node_labels():
    # Arrange
    graph = nx.Graph()
    graph.add_node(1, label="Node 1")
    graph.add_node(2, label="Node 2")
    graph.add_edge(1, 2)
    builder = DiagramBuilder()

    # Act
    diagram = builder.build(graph)

    # Assert
    assert "A(Node 1)" in diagram
    assert "B(Node 2)" in diagram
    assert "A --> B" in diagram


def test_build_graph_with_edge_labels():
    # Arrange
    graph = nx.Graph()
    graph.add_edge(1, 2, label="Edge 1-2")
    builder = DiagramBuilder()

    # Act
    diagram = builder.build(graph)

    # Assert
    assert "A -->|Edge 1-2| B" in diagram


def test_build_graph_without_edge_labels():
    # Arrange
    graph = nx.Graph()
    graph.add_edge(1, 2, label="Edge 1-2")
    builder = DiagramBuilder()

    # Act
    diagram = builder.build(graph, with_edge_labels=False)

    # Assert
    assert "A --> B" in diagram
    assert "|Edge 1-2|" not in diagram


def test_build_graph_with_node_colors():
    # Arrange
    graph = nx.Graph()
    graph.add_node(1, color="#FF0000")
    graph.add_node(2, color="#00FF00")
    graph.add_edge(1, 2)
    builder = DiagramBuilder()

    # Act
    diagram = builder.build(graph)

    # Assert
    assert "style A fill:#FF0000, color:#ffffff" in diagram
    assert "style B fill:#00FF00, color:#ffffff" in diagram


def test_build_graph_with_custom_node_shape():
    # Arrange
    graph = nx.Graph()
    graph.add_node(1)
    graph.add_edge(1, 1)
    builder = DiagramBuilder(node_shape=DiagramNodeShape.RECTANGLE)

    # Act
    diagram = builder.build(graph)

    # Assert
    assert "A[1]" in diagram


def test_build_graph_with_custom_orientation():
    # Arrange
    graph = nx.Graph()
    graph.add_node(1)
    graph.add_edge(1, 1)
    builder = DiagramBuilder(orientation=DiagramOrientation.TOP_DOWN)

    # Act
    diagram = builder.build(graph)

    # Assert
    assert "graph TD" in diagram


def test_build_graph_with_custom_layout():
    # Arrange
    graph = nx.Graph()
    graph.add_node(1)
    graph.add_edge(1, 1)
    builder = DiagramBuilder(layout='elk')

    # Act
    diagram = builder.build(graph)

    # Assert
    assert "layout: elk" in diagram


def test_build_graph_with_custom_look():
    # Arrange
    graph = nx.Graph()
    graph.add_node(1)
    graph.add_edge(1, 1)
    builder = DiagramBuilder(look='neo')

    # Act
    diagram = builder.build(graph)

    # Assert
    assert "look: neo" in diagram


def test_build_graph_with_custom_theme():
    # Arrange
    graph = nx.Graph()
    graph.add_node(1)
    graph.add_edge(1, 1)
    builder = DiagramBuilder(theme='neutral')

    # Act
    diagram = builder.build(graph)

    # Assert
    assert "theme: neutral" in diagram


def test_build_graph_with_title():
    # Arrange
    graph = nx.Graph(name="My Graph")
    graph.add_node(1)
    graph.add_edge(1, 1)
    builder = DiagramBuilder()

    # Act
    diagram = builder.build(graph)

    # Assert
    assert "title: My Graph" in diagram


def test_build_graph_without_title():
    # Arrange
    graph = nx.Graph()
    graph.add_node(1)
    graph.add_edge(1, 1)
    builder = DiagramBuilder()

    # Act
    diagram = builder.build(graph)

    # Assert
    assert "title" not in diagram


def test_build_graph_with_title_overwrite():
    # Arrange
    graph = nx.Graph(name="My Graph")
    graph.add_node(1)
    graph.add_edge(1, 1)
    builder = DiagramBuilder()

    # Act
    diagram = builder.build(graph, title="Custom Title")

    # Assert
    assert "title: Custom Title" in diagram
    assert "title: My Graph" not in diagram


def test_build_graph_with_empty_title_overwrite():
    # Arrange
    graph = nx.Graph(name="My Graph")
    graph.add_node(1)
    graph.add_edge(1, 1)
    builder = DiagramBuilder()

    # Act
    diagram = builder.build(graph, title="")

    # Assert
    assert "title" not in diagram


def test_edge_label_with_label():
    # Arrange
    data = {"label": "Edge Label"}

    # Act
    result = _edge_label(data)

    # Assert
    assert result == "|Edge Label|"


def test_edge_label_without_label():
    # Arrange
    data = {}

    # Act
    result = _edge_label(data)

    # Assert
    assert result == ""


@pytest.mark.parametrize(
    ('color', 'expected_contrast'),
    [
        ("#FFFFFF", "#000000"),
        ("#000000", "#ffffff"),
        ("#FF0000", "#ffffff"),
        ("#00FF00", "#ffffff"),
        ("#0000FF", "#ffffff"),
        ("#ABCDEF", "#000000"),
        ("#123456", "#ffffff"),
    ],
)
def test_contrast_color_valid_hex(color, expected_contrast):
    # Arrange (done by parametrizing)
    # Act
    result = _contrast_color(color)

    # Assert
    assert result == expected_contrast


@pytest.mark.parametrize(
    "color",
    ["FFFFFF", "#FFFFF", "#GGGGGG", 123456, None],
)
def test_contrast_color_invalid_hex(color):
    # Arrange (done by parametrizing)
    # Act & Assert
    with pytest.raises(ValueError):  # noqa: PT011
        _contrast_color(color)


def test_node_style_with_color():
    # Arrange
    data = {"color": "#FF0000"}

    # Act
    result = _node_style("1", data)

    # Assert
    assert result == "\nstyle 1 fill:#FF0000, color:#ffffff"


def test_node_style_without_color():
    # Arrange
    data = {}

    # Act
    result = _node_style("1", data)

    # Assert
    assert result == ""


def test_graph_title_with_name():
    # Arrange
    graph = nx.Graph(name="My Graph")

    # Act
    result = _graph_title(graph)

    # Assert
    assert result == "title: My Graph\n"


def test_graph_title_without_name():
    # Arrange
    graph = nx.Graph()

    # Act
    result = _graph_title(graph)

    # Assert
    assert result == ""
